import functools as ft
import inspect
from typing import Any, Callable, Dict, Union

import jax
import jax.numpy as jnp

from .compile_utils import (
    compile_cache,
    get_fun_names,
    hashable_combine,
    hashable_partition,
    Static,
)
from .custom_types import BoolAxisSpec, PyTree, ResolvedBoolAxisSpec, sentinel
from .doc_utils import doc_strip_annotations
from .filters import combine, filter, is_array, is_array_like, partition
from .module import Module, module_update_wrapper


ResolvedMapAxisSpec = Union[None, int]
MapAxisSpec = Union[ResolvedMapAxisSpec, Callable[[Any], ResolvedMapAxisSpec]]
#
ResolvedAxisSpec = Union[ResolvedBoolAxisSpec, ResolvedMapAxisSpec]
AxisSpec = Union[ResolvedAxisSpec, Callable[[Any], ResolvedAxisSpec]]


def _is_none(x: Any) -> bool:
    return x is None


def _resolve_axis(axis_spec: AxisSpec, elem: Any) -> PyTree[ResolvedAxisSpec]:
    if axis_spec is None or isinstance(axis_spec, (bool, int)):
        return axis_spec
    if callable(axis_spec):
        return jax.tree_map(axis_spec, elem)
    else:
        raise ValueError(
            "`in_axes` and `out_axes` must consist of None, bools, ints, and callables only."
        )


def _resolve_axes(
    pytree: PyTree[Any], axes_spec: PyTree[AxisSpec]
) -> PyTree[ResolvedAxisSpec]:
    return jax.tree_map(_resolve_axis, axes_spec, pytree, is_leaf=_is_none)


def _jit_axis(axis: ResolvedAxisSpec) -> BoolAxisSpec:  # not necessarily resolved
    if isinstance(axis, bool):
        return axis
    elif isinstance(axis, int):
        return True
    elif axis is None:
        return is_array
    else:
        assert False


def _jit_axes(axes: PyTree[ResolvedAxisSpec]) -> PyTree[BoolAxisSpec]:
    return jax.tree_map(_jit_axis, axes, is_leaf=_is_none)


def _map_axis(axis: ResolvedAxisSpec) -> ResolvedMapAxisSpec:
    if isinstance(axis, bool):
        return None
    elif isinstance(axis, int):
        return axis
    elif axis is None:
        return None
    else:
        assert False


def _map_axes(axes: PyTree[ResolvedAxisSpec]) -> PyTree[ResolvedMapAxisSpec]:
    return jax.tree_map(_map_axis, axes, is_leaf=_is_none)


def _zero_if_array_else_none(x: Any) -> ResolvedMapAxisSpec:
    return 0 if is_array(x) else None


def _check_map_out_axis(name, max_out_size, x):
    if isinstance(x, bool) or x is None:
        pass
    elif isinstance(x, int):
        if x < -max_out_size or x >= max_out_size:
            raise ValueError(
                f"integers in filter_{name}map(..., out=...) must correspond to a "
                "dimension of the output array"
            )
    else:
        raise ValueError(
            f"filter_{name}map(..., out=...) must contain only integers and Nones"
        )


def _eval_and_return_arraylike(static, dynamic):
    fun, args, kwargs = combine(static, dynamic)
    out = fun(*args, **kwargs)
    return filter(out, is_array_like)


def _eval_shape(leaves, treedef, axis_name, axis_size):
    static, struct, in_axes = jax.tree_unflatten(treedef, leaves)
    fn = ft.partial(_eval_and_return_arraylike, static)
    # To get just the number of dimensions we don't need to set out_axes; the batch
    # axis will always go at the front but we don't care.
    mapkwargs = {}
    if axis_name is not sentinel:
        mapkwargs["axis_name"] = axis_name
    if axis_size is not sentinel:
        mapkwargs["axis_size"] = axis_size
    fn = jax.vmap(fn, in_axes=(in_axes,), **mapkwargs)
    struct_out = jax.eval_shape(fn, struct)
    return jax.tree_util.tree_reduce(lambda x, y: max(x, y.ndim), struct_out, 0)


_eval_shape_cache = ft.lru_cache(maxsize=None)(_eval_shape)


# Calculates the maximum `ndim` over all array outputs of `fun`.
#
# Note the asymmetry between `is_array` filtering on the input and `is_array_like`
# filtering on the output. We don't want to trace non-arrays as they might be boolean
# flags etc. (And if the user is using non-arrays in array-like ways then that's fine.)
# Hence `is_array` on the input. However on the output then we may-or-may-not be
# batching non-array outputs (i.e. broadcasting them along the batch dimension). We
# play it safe and assume we broadcast all of them, hence `is_array_like`. (And any
# that we broadcast unnecessarily mean we might just get a return value of `1` rather
# than `0` in some edge cases, which is fine since that will raise an error where it's
# actually used in `out_axes`.)
def _get_max_out_size(fun, args, kwargs, in_axes, mapkwargs, cache) -> int:
    dynamic, static = partition((fun, args, kwargs), is_array)
    try:
        axis_name = mapkwargs["axis_name"]
    except KeyError:
        axis_name = sentinel
    try:
        axis_size = mapkwargs["axis_size"]
    except KeyError:
        axis_size = sentinel
    struct = jax.tree_map(lambda x: jax.ShapeDtypeStruct(x.shape, x.dtype), dynamic)
    if cache:
        _eval_shape_fn = _eval_shape_cache
    else:
        _eval_shape_fn = _eval_shape
    leaves, treedef = jax.tree_flatten((static, struct, in_axes))
    leaves = tuple(leaves)
    return _eval_shape_fn(leaves, treedef, axis_name, axis_size)


class _VmapWrapper(Module):
    _signature: inspect.Signature
    _fun: Callable
    _default: AxisSpec
    _fn: PyTree[AxisSpec]
    _args: PyTree[AxisSpec]
    _kwargs: PyTree[AxisSpec]
    _out: PyTree[AxisSpec]
    _vmapkwargs: Dict[str, Any]

    def __call__(__self, *args, **kwargs):
        def _fun_wrapper(_fun, _args, _kwargs):
            _out = _fun(*_args, **_kwargs)

            _out_axes = _resolve_axes(_out, __self._out)
            _out_axes = _map_axes(_out_axes)
            jax.tree_map(ft.partial(_check_map_out_axis, "v", max_out_size), _out_axes)
            _vmapd = []
            for i in range(-max_out_size, max_out_size):
                _i_axes = jax.tree_map(lambda a: a == i, _out_axes)
                _vmapd.append(filter(_out, _i_axes))
            _none_axes = jax.tree_map(_is_none, _out_axes, is_leaf=_is_none)
            _nonvmapd = filter(_out, _none_axes)
            return _vmapd, Static(_nonvmapd)

        bound = __self._signature.bind(*args, **kwargs)
        del args, kwargs
        bound.apply_defaults()

        _args = __self._args + (__self._default,) * (
            len(bound.args) - len(__self._args)
        )
        _kwargs = {
            key: __self._kwargs.get(key, __self._default) for key in bound.kwargs
        }
        in_axes = _resolve_axes(
            (__self._fun, bound.args, bound.kwargs), (__self._fn, _args, _kwargs)
        )
        del _args, _kwargs
        in_axes = _map_axes(in_axes)

        max_out_size = _get_max_out_size(
            __self._fun,
            bound.args,
            bound.kwargs,
            in_axes,
            __self._vmapkwargs,
            cache=False,
        )
        out_axes = (list(range(-max_out_size, max_out_size)), None)
        vmapd, nonvmapd = jax.vmap(
            _fun_wrapper, in_axes=in_axes, out_axes=out_axes, **__self._vmapkwargs
        )(__self._fun, bound.args, bound.kwargs)
        return combine(*vmapd, nonvmapd.value)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return jax.tree_util.Partial(self, instance)


# Note the use of AxisSpec rather than MapAxisSpec.
# This is to support seamlessly switching out filter_pmap for filter_vmap.
@doc_strip_annotations
def filter_vmap(
    fun: Callable = sentinel,
    *,
    default: AxisSpec = _zero_if_array_else_none,
    fn: PyTree[AxisSpec] = None,
    args: PyTree[AxisSpec] = (),
    kwargs: PyTree[AxisSpec] = None,
    out: PyTree[AxisSpec] = _zero_if_array_else_none,
    **vmapkwargs,
) -> Callable:
    """Wraps together [`equinox.partition`][] and `jax.vmap`.

    !!! info

        By default, all JAX arrays are vectorised down their leading axis (i.e. axis
        index 0), and all other types are not vectorised.

    **Arguments:**

    In each of the following cases, then `int` indicates an array axis to vectorise
    over, `None` indicates that an argument should be broadcast (not vectorised
    over), and functions `Leaf -> Union[None, int]` are mapped and evaluated on every
    leaf of their subtree. `None` should be used for non-JAX-array arguments.

    - `fun` is a pure function to vectorise.
    - `default` should be a `Union[None, int]` or a function
        `Leaf -> Union[None, int]`, and is applied by default to every argument and
        keyword argument to `fun`.
    - `args` is an optional per-argument override for `default`, and should be a tuple
        of PyTrees with leaves that are either `Union[None, int]`s or functions
        `Leaf -> Union[None, int]`. The PyTree structures should be prefixes of the
        corresponding input to `fun`.
    - `kwargs` is an optional per-keyword-argument override for `default` and should be
        a dictionary, whose keys are the names of arguments to `fun`, and whose values
        are PyTrees with leaves that are either `Union[None, int]`s or functions
        `Leaf -> Union[None, int]`. The PyTree structures should be prefixes of the
        corresponding input to `fun`.
    - `out` should be a PyTree with leaves that are either `Union[None, int]`s or
        functions `Leaf -> Union[None, int]`. The PyTree structure should be a prefix
        of the output of `fun`.
    - `fn` should be a PyTree with leaves that are either `Union[None, int]`s or
        functions `Leaf -> Union[None, int]`. The PyTree structure should be a prefix of
        `fun` itself. (Note that `fun` may be any callable, e.g. a bound method, or a
        class implementing `__call__`, and doesn't have to be a normal Python function.)
    - `**vmapkwargs` are any other keyword arguments to `jax.vmap`.

    When `args`, `kwargs`, `out`, `fn` are prefixes of the corresponding input, their
    value will be mapped over the input PyTree.

    **Returns:**

    The vectorised version of `fun`.

    !!! info

        In fact, besides `None`, `int` and `Leaf -> Union[None, int]`: boolean
        types are also supported and treated identically to `None`. This is to support
        seamlessly switching between [`equinox.filter_pmap`][] and
        [`equinox.filter_vmap`][] if desired.

    !!! example

        ```python
        import equinox as eqx
        import jax.numpy as jnp

        @eqx.filter_vmap
        def f(x, y):
            return x + y

        @eqx.filter_vmap(kwargs=dict(x=1))
        def g(x, y):
            return x + y

        @eqx.filter_vmap(args=(None,))
        def h(x, y):
            return x + y

        f(jnp.array([1, 2]), jnp.array([3, 4]))  # both args vectorised down axis 0
        f(jnp.array([1, 2]), 3)  # first arg vectorised down axis 0
                                 # second arg broadcasted

        g(jnp.array([[1, 2]]), jnp.array([3, 4]))  # first arg vectorised down axis 1
                                                   # second arg vectorised down axis 0

        h(jnp.array(1), jnp.array([2, 3]))  # first arg broadcasted
                                            # second arg vectorised down axis 0
        ```

    !!! example

        `filter_vmap` can be used to easily create ensembles of models. For example, here's an
        ensemble of eight MLPs:

        ```python
        import equinox as eqx
        import jax.random as jr

        key = jr.PRNGKey(0)
        keys = jr.split(key, 8)

        # Create an ensemble of models

        @eqx.filter_vmap
        def make_ensemble(key):
            return eqx.nn.MLP(2, 2, 2, 2, key=key)

        mlp_ensemble = make_ensemble(keys)

        # Evaluate each member of the ensemble on the same data

        @eqx.filter_vmap(kwargs=dict(x=None))
        def evaluate_ensemble(model, x):
            return model(x)

        evaluate_ensemble(mlp_ensemble, jr.normal(key, (2,)))

        # Evaluate each member of the ensemble on different data

        @eqx.filter_vmap
        def evaluate_per_ensemble(model, x):
            return model(x)

        evaluate_per_ensemble(mlp_ensemble, jr.normal(key, (8, 2)))
        ```

        Here, `make_ensemble` works because [`equinox.nn.MLP`][] is a PyTree, and so it
        is a valid output from a `filter_vmap`. This PyTree includes some JAX arrays
        (the weights and biases) and some non-JAX-arrays (e.g. activation functions).
        `filter_vmap` will vectorise the JAX arrays (with separate weights for each
        member of the ensemble) whilst leaving the non-JAX-arrays alone.

        Note that as the weights in `mlp_ensemble` now have a leading batch dimension
        -- that the weights of `eqx.nn.MLP` instances do not typically have -- then it
        cannot be called directly. It must instead be passed back into a vectorised
        region to be called.
    """

    if fun is sentinel:
        return ft.partial(
            filter_vmap,
            default=default,
            fn=fn,
            args=args,
            kwargs=kwargs,
            out=out,
            **vmapkwargs,
        )

    if kwargs is None:
        kwargs = {}

    signature = inspect.signature(fun)

    signature_default = signature.replace(
        parameters=[
            p
            if p.kind
            in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD)
            else p.replace(default=default)
            for p in signature.parameters.values()
        ]
    )
    bound = signature_default.bind_partial(*args, **kwargs)
    del args, kwargs
    bound.apply_defaults()

    vmap_wrapper = _VmapWrapper(
        _signature=signature,
        _fun=fun,
        _default=default,
        _fn=fn,
        _args=bound.args,
        _kwargs=bound.kwargs,
        _out=out,
        _vmapkwargs=vmapkwargs,
    )
    return module_update_wrapper(vmap_wrapper, fun)


@compile_cache
def _filter_pmap_cache(fun_names, map_in_axes, max_out_size, **pmapkwargs):
    def fun_wrapped(_dynamic, _static_leaves, _static_treedef, _out_axes):
        _fun, _args, _kwargs, _maybe_dummy = hashable_combine(
            _dynamic, _static_leaves, _static_treedef
        )
        del _maybe_dummy
        _out = _fun(*_args, **_kwargs)

        _out_axes = _resolve_axes(_out, _out_axes)
        _map_out_axes = _map_axes(_out_axes)
        jax.tree_map(ft.partial(_check_map_out_axis, "p", max_out_size), _map_out_axes)
        _jit_out_axes = _jit_axes(_out_axes)

        _dynamic, _static_leaves, _static_treedef = hashable_partition(
            _out, _jit_out_axes
        )

        _pmapd = []
        for i in range(-max_out_size, max_out_size):
            _i_axes = jax.tree_map(lambda a: a == i, _map_out_axes)
            _pmapd.append(filter(_dynamic, _i_axes))
        _none_axes = jax.tree_map(_is_none, _map_out_axes, is_leaf=_is_none)
        _nonpmapd = filter(_dynamic, _none_axes)

        return (
            _pmapd,
            _nonpmapd,
            Static(_static_leaves),
            Static(_static_treedef),
        )

    fun_name, fun_qualname = fun_names
    fun_wrapped.__name__ = fun_name
    fun_wrapped.__qualname__ = fun_qualname

    map_out_axes = (list(range(-max_out_size, max_out_size)), None, None, None)

    return jax.pmap(
        fun_wrapped,
        in_axes=(map_in_axes, None, None, None),
        out_axes=map_out_axes,
        static_broadcasted_argnums=(1, 2, 3),
        **pmapkwargs,
    )


class _PmapWrapper(Module):
    _signature: inspect.Signature
    _fun: Callable
    _default: AxisSpec
    _fn: PyTree[AxisSpec]
    _args: PyTree[AxisSpec]
    _kwargs: PyTree[AxisSpec]
    _out: PyTree[AxisSpec]
    _pmapkwargs: Dict[str, Any]

    def _call(self, is_lower, args, kwargs):
        bound = self._signature.bind(*args, **kwargs)
        del args, kwargs
        bound.apply_defaults()

        _args = self._args + (self._default,) * (len(bound.args) - len(self._args))
        _kwargs = {key: self._kwargs.get(key, self._default) for key in bound.kwargs}
        try:
            axis_size = self._pmapkwargs["axis_size"]
        except KeyError:
            maybe_dummy = 0  # hashable non-array object
        else:
            # Work around JAX issue #9252
            maybe_dummy = jnp.empty(axis_size)
        in_axes = _resolve_axes(
            (self._fun, bound.args, bound.kwargs, maybe_dummy),
            (self._fn, _args, _kwargs, _zero_if_array_else_none),
        )
        del _args, _kwargs
        jit_in_axes = _jit_axes(in_axes)
        map_in_axes = _map_axes(in_axes)

        map_in_axes_no_dummy = map_in_axes[:-1]
        max_out_size = _get_max_out_size(
            self._fun,
            bound.args,
            bound.kwargs,
            map_in_axes_no_dummy,
            self._pmapkwargs,
            cache=True,
        )

        cached = _filter_pmap_cache(
            get_fun_names(self._fun),
            map_in_axes=map_in_axes,
            max_out_size=max_out_size,
            **self._pmapkwargs,
        )

        dynamic, static_leaves, static_treedef = hashable_partition(
            (self._fun, bound.args, bound.kwargs, maybe_dummy), jit_in_axes
        )
        if is_lower:
            return cached.lower(dynamic, static_leaves, static_treedef, self._out)
        else:
            (
                pmapd,
                dynamic_nonpmapd,
                static_nonpmapd_leaves,
                static_nonpmapd_treedef,
            ) = cached(dynamic, static_leaves, static_treedef, self._out)
            nonpmapd = hashable_combine(
                dynamic_nonpmapd,
                static_nonpmapd_leaves.value,
                static_nonpmapd_treedef.value,
            )
            return combine(*pmapd, nonpmapd)

    def __call__(__self, *args, **kwargs):
        return __self._call(False, args, kwargs)

    def lower(__self, *args, **kwargs):
        return __self._call(True, args, kwargs)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return jax.tree_util.Partial(self, instance)


@doc_strip_annotations
def filter_pmap(
    fun: Callable = sentinel,
    *,
    default: AxisSpec = _zero_if_array_else_none,
    fn: PyTree[AxisSpec] = None,
    args: PyTree[AxisSpec] = (),
    kwargs: PyTree[AxisSpec] = None,
    out: PyTree[AxisSpec] = _zero_if_array_else_none,
    **pmapkwargs,
) -> Callable:
    """Wraps together [`equinox.partition`][] and `jax.pmap`.

    !!! info

        By default, the computation is parallelised by splitting all JAX arrays down
        their leading axis (i.e. axis index 0), and broadcasting all other types to
        each replica.

    **Arguments:**

    In each of the following cases, then `int` indicates an array axis to split down,
    `None` indicates that an argument should be broadcast to each device (not split
    across devices), and functions `Leaf -> Union[None, bool, int]` are mapped and
    evaluated on every leaf of their subtree.

    Note that `jax.pmap`, and thus `equinox.filter_pmap`, also JIT-compile their
    function in the same way as `jax.jit`. By default, all JAX arrays are traced and
    all other arrays are treated as static inputs. This may be controlled explicitly
    -- instead of just passing `None` -- by passing either `True` (traced) or
    `False` (static).

    `None`, `False` and `True` should be used for non-JAX-array arguments.

    - `fun` is a pure function to parallelise.
    - `default` should be a `Union[None, bool, int]` or a function
        `Leaf -> Union[None, bool, int]`, and is applied by default to every argument
        and keyword argument to `fun`.
    - `args` is an optional per-argument override for `default`, and should be a tuple
        of PyTrees with leaves that are either `Union[None, bool, int]`s or functions
        `Leaf -> Union[None, bool, int]`. The PyTree structures should be prefixes of
        the corresponding input to `fun`.
    - `kwargs` is an optional per-keyword-argument override for `default` and should be
        a dictionary, whose keys are the names of arguments to `fun`, and whose values
        are PyTrees with leaves that are either `Union[None, bool, int]`s or functions
        `Leaf -> Union[None, bool, int]`. The PyTree structures should be prefixes of
        the corresponding input to `fun`.
    - `out` should be a PyTree with leaves that are either `Union[None, bool, int]`s or
        functions `Leaf -> Union[None, bool, int]`. The PyTree structure should be a
        prefix of the output of `fun`. `True` indicates a tracer, `False` indicates any
        auxiliary information to return.
    - `fn` should be a PyTree with leaves that are either `Union[None, bool, int]`s or
        functions `Leaf -> Union[None, bool, int]`. The PyTree structure should be a
        prefix of `fun` itself. (Note that `fun` may be any callable, e.g. a bound
        method, or a class implementing `__call__`, and doesn't have to be a normal
        Python function.)
    - `**pmapkwargs` are any other keyword arguments to `jax.pmap`.

    When `args`, `kwargs`, `out`, `fn` are prefixes of the corresponding input, their
    value will be mapped over the input PyTree.

    **Returns:**

    The parallelised version of `fun`.

    !!! example

        ```python
        import equinox as eqx
        import jax.numpy as jnp

        @eqx.filter_pmap
        def f(x, y):
            return x + y

        @eqx.filter_pmap(kwargs=dict(x=1))
        def g(x, y):
            return x + y

        @eqx.filter_pmap(args=(None,))
        def h(x, y):
            return x + y

        @eqx.filter_pmap
        def apply(fun, x):
            return fun(x)

        f(jnp.array([1, 2]), jnp.array([3, 4]))  # both args split down axis 0
        f(jnp.array([1, 2]), 3)  # first arg split down axis 0
                                 # second arg broadcasted

        g(jnp.array([[1, 2]]), jnp.array([3, 4]))  # first arg split down axis 1
                                                   # second arg split down axis 0

        h(jnp.array(1), jnp.array([2, 3]))  # first arg broadcasted
                                            # second arg split down axis 0

        apply(lambda x: x + 1, jnp.array([2, 3]))  # first arg broadcasted (as it's not
                                                   # a JAX array)
                                                   # second arg split down axis 0
        ```
    """

    if fun is sentinel:
        return ft.partial(
            filter_pmap,
            default=default,
            fn=fn,
            args=args,
            kwargs=kwargs,
            out=out,
            **pmapkwargs,
        )

    if kwargs is None:
        kwargs = {}

    signature = inspect.signature(fun)

    signature_default = signature.replace(
        parameters=[
            p
            if p.kind
            in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD)
            else p.replace(default=default)
            for p in signature.parameters.values()
        ]
    )
    bound = signature_default.bind_partial(*args, **kwargs)
    del args, kwargs
    bound.apply_defaults()

    pmap_wrapper = _PmapWrapper(
        _signature=signature,
        _fun=fun,
        _default=default,
        _fn=fn,
        _args=bound.args,
        _kwargs=bound.kwargs,
        _out=out,
        _pmapkwargs=pmapkwargs,
    )

    return module_update_wrapper(pmap_wrapper, fun)



import dataclasses
import functools as ft
import types
from typing import Any, Callable, Dict, List, NamedTuple, Tuple

import jax
import jax._src.pretty_printer as pp
import jax.numpy as jnp
import numpy as np

from .custom_types import Array, PyTree


Dataclass = Any
PrettyPrintAble = PyTree


_comma_sep = pp.concat([pp.text(","), pp.brk()])


def _nest(n: int, doc: pp.Doc) -> pp.Doc:
    return pp.nest(n, pp.concat([pp.brk(""), doc]))


def _pformat_list(obj: List, **kwargs) -> pp.Doc:
    indent = kwargs["indent"]
    return pp.group(
        pp.concat(
            [
                pp.text("["),
                _nest(
                    indent, pp.join(_comma_sep, [_pformat(x, **kwargs) for x in obj])
                ),
                pp.brk(""),
                pp.text("]"),
            ]
        )
    )


def _pformat_tuple(obj: Tuple, **kwargs) -> pp.Doc:
    indent = kwargs["indent"]
    if len(obj) == 1:
        entries = pp.concat([_pformat(obj[0], **kwargs), pp.text(",")])
    else:
        entries = pp.join(_comma_sep, [_pformat(x, **kwargs) for x in obj])
    return pp.group(
        pp.concat([pp.text("("), _nest(indent, entries), pp.brk(""), pp.text(")")])
    )


def _dict_entry(key: PrettyPrintAble, value: PrettyPrintAble, **kwargs) -> pp.Doc:
    return pp.concat(
        [_pformat(key, **kwargs), pp.text(":"), pp.brk(), _pformat(value, **kwargs)]
    )


def _pformat_dict(obj: Dict, **kwargs) -> pp.Doc:
    indent = kwargs["indent"]
    entries = [_dict_entry(key, value, **kwargs) for key, value in obj.items()]
    return pp.group(
        pp.concat(
            [
                pp.text("{"),
                _nest(indent, pp.join(_comma_sep, entries)),
                pp.brk(""),
                pp.text("}"),
            ]
        )
    )


def _named_entry(name: str, value: Any, **kwargs) -> pp.Doc:
    return pp.concat([pp.text(name), pp.text("="), _pformat(value, **kwargs)])


def _pformat_namedtuple(obj: NamedTuple, **kwargs) -> pp.Doc:
    indent = kwargs["indent"]
    entries = [_named_entry(name, getattr(obj, name), **kwargs) for name in obj._fields]
    return pp.group(
        pp.concat(
            [
                pp.text(obj.__class__.__name__),
                pp.text("("),
                _nest(indent, pp.join(_comma_sep, entries)),
                pp.brk(""),
                pp.text(")"),
            ]
        )
    )


def _pformat_dataclass(obj: Dataclass, **kwargs) -> pp.Doc:
    indent = kwargs["indent"]
    entries = [
        _named_entry(f.name, getattr(obj, f.name), **kwargs)
        for f in dataclasses.fields(obj)
        if f.repr
    ]
    return pp.group(
        pp.concat(
            [
                pp.text(obj.__class__.__name__),
                pp.text("("),
                _nest(indent, pp.join(_comma_sep, entries)),
                pp.brk(""),
                pp.text(")"),
            ]
        )
    )


def _pformat_array(obj: Array, **kwargs) -> pp.Doc:
    short = kwargs["short_arrays"]
    if short:
        dtype_str = (
            obj.dtype.name.replace("float", "f")
            .replace("uint", "u")
            .replace("int", "i")
            .replace("complex", "c")
        )
        shape_str = ",".join(map(str, obj.shape))
        backend = "(numpy)" if isinstance(obj, np.ndarray) else ""
        return pp.text(f"{dtype_str}[{shape_str}]{backend}")
    else:
        return pp.text(repr(obj))


def _pformat_function(obj: types.FunctionType, **kwargs) -> pp.Doc:
    if kwargs.get("wrapped", False):
        fn = "wrapped function"
    else:
        fn = "function"
    return pp.text(f"<{fn} {obj.__name__}>")


# TODO: offer user-customisable override mechanism if they wish.
# Ideally this would be something tied to being a PyTree.
def _pformat(obj: PrettyPrintAble, **kwargs) -> pp.Doc:
    follow_wrapped = kwargs["follow_wrapped"]
    truncate_leaf = kwargs["truncate_leaf"]
    if truncate_leaf(obj):
        return pp.text(f"{type(obj).__name__}(...)")
    elif dataclasses.is_dataclass(obj):
        return _pformat_dataclass(obj, **kwargs)
    elif isinstance(obj, list):
        return _pformat_list(obj, **kwargs)
    elif isinstance(obj, dict):
        return _pformat_dict(obj, **kwargs)
    elif isinstance(obj, tuple):
        if hasattr(obj, "_fields"):
            return _pformat_namedtuple(obj, **kwargs)
        else:
            return _pformat_tuple(obj, **kwargs)
    elif isinstance(obj, (np.ndarray, jnp.ndarray)):
        return _pformat_array(obj, **kwargs)
    elif isinstance(obj, (jax.custom_jvp, jax.custom_vjp)):
        return _pformat(obj.__wrapped__, **kwargs)
    elif hasattr(obj, "__wrapped__") and follow_wrapped:
        kwargs["wrapped"] = True
        return _pformat(obj.__wrapped__, **kwargs)
    elif isinstance(obj, ft.partial) and follow_wrapped:
        kwargs["wrapped"] = True
        return _pformat(obj.func, **kwargs)
    elif isinstance(obj, types.FunctionType):
        return _pformat_function(obj, **kwargs)
    else:  # int, str, float, complex, bool, etc.
        return pp.text(repr(obj))


def _false(_):
    return False


def tree_pformat(
    pytree: PrettyPrintAble,
    width: int = 80,
    indent: int = 2,
    short_arrays: bool = True,
    follow_wrapped: bool = True,
    truncate_leaf: Callable[[PrettyPrintAble], bool] = _false,
) -> str:
    """Pretty-formats a PyTree as a string, whilst abbreviating JAX arrays.

    (This is the function used in `__repr__` of [`equinox.Module`][].)

    All JAX arrays in the PyTree are condensed down to a short string representation
    of their dtype and shape.

    !!! example

        A 32-bit floating-point JAX array of shape `(3, 4)` is printed as `f32[3,4]`.

    **Arguments:**

    - `pytree`: The PyTree to pretty-format.
    - `width`: The desired maximum number of characters per line of output. If a
        structure cannot be formatted within this constraint then a best effort will
        be made.
    - `indent`: The amount of indentation each nesting level.
    - `short_arrays`: Toggles the abbreviation of JAX arrays.
    - `follow_wrapped`: Whether to unwrap `functools.partial` and `functools.wraps`.
    - `truncate_leaf`: A function `Any -> bool`. Applied to all nodes in the PyTree;
        all truthy nodes will be truncated to just `f"{type(node).__name__}(...)"`.

    **Returns:**

    A string.
    """

    return _pformat(
        pytree,
        indent=indent,
        short_arrays=short_arrays,
        follow_wrapped=follow_wrapped,
        truncate_leaf=truncate_leaf,
    ).format(width=width)



from typing import Callable, Optional, Sequence, Tuple, Union

import jax.lax as lax
import jax.numpy as jnp
import jax.random
import numpy as np

from ..custom_types import Array
from ..module import Module, static_field


class Pool(Module):
    """General N-dimensional downsampling over a sliding window."""

    init: Union[int, float, Array]
    operation: Callable[[Array, Array], Array]
    num_spatial_dims: int = static_field()
    kernel_size: Union[int, Sequence[int]] = static_field()
    stride: Union[int, Sequence[int]] = static_field()
    padding: Union[int, Sequence[int], Sequence[Tuple[int, int]]] = static_field()

    def __init__(
        self,
        init: Union[int, float, Array],
        operation: Callable[[Array, Array], Array],
        num_spatial_dims: int,
        kernel_size: Union[int, Sequence[int]],
        stride: Union[int, Sequence[int]] = 1,
        padding: Union[int, Sequence[int], Sequence[Tuple[int, int]]] = 0,
        **kwargs,
    ):
        """**Arguments:**

        - `init`: The initial value for the reduction.
        - `operation`: The operation applied to the inputs of each window.
        - `num_spatial_dims`: The number of spatial dimensions.
        - `kernel_size`: The size of the convolutional kernel.
        - `stride`: The stride of the convolution.
        - `padding`: The amount of padding to apply before and after each
            spatial dimension.

        !!! info

            In order for `Pool` to be differentiable, `operation(init, x) == x` needs to
            be true for all finite `x`. For further details see
            [https://www.tensorflow.org/xla/operation_semantics#reducewindow](https://www.tensorflow.org/xla/operation_semantics#reducewindow)
            and [https://github.com/google/jax/issues/7718](https://github.com/google/jax/issues/7718).
        """
        super().__init__(**kwargs)

        self.operation = operation
        self.init = init
        self.num_spatial_dims = num_spatial_dims

        if isinstance(kernel_size, int):
            self.kernel_size = (kernel_size,) * num_spatial_dims
        elif isinstance(kernel_size, Sequence):
            self.kernel_size = kernel_size
        else:
            raise ValueError(
                "`kernel_size` must either be an int or tuple of length "
                f"{num_spatial_dims} containing ints."
            )

        if isinstance(stride, int):
            self.stride = (stride,) * num_spatial_dims
        elif isinstance(stride, Sequence):
            self.stride = stride
        else:
            raise ValueError(
                "`stride` must either be an int or tuple of length "
                f"{num_spatial_dims} containing ints."
            )

        if isinstance(padding, int):
            self.padding = tuple((padding, padding) for _ in range(num_spatial_dims))
        elif isinstance(padding, Sequence) and len(padding) == num_spatial_dims:
            if all(isinstance(element, Sequence) for element in padding):
                self.padding = padding
            else:
                self.padding = tuple((p, p) for p in padding)
        else:
            raise ValueError(
                "`padding` must either be an int or tuple of length "
                f"{num_spatial_dims} containing ints or tuples of length 2."
            )

    def __call__(
        self, x: Array, *, key: Optional["jax.random.PRNGKey"] = None
    ) -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array of shape `(channels, dim_1, ..., dim_N)`, where
            `N = num_spatial_dims`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(channels, new_dim_1, ..., new_dim_N)`.
        """
        assert len(x.shape) == self.num_spatial_dims + 1, (
            f"Input should have {self.num_spatial_dims} spatial dimensions, "
            f"but input has shape {x.shape}"
        )

        x = jnp.moveaxis(x, 0, -1)
        x = jnp.expand_dims(x, axis=0)
        x = lax.reduce_window(
            x,
            self.init,
            self.operation,
            (1,) + self.kernel_size + (1,),
            (1,) + self.stride + (1,),
            ((0, 0),) + self.padding + ((0, 0),),
        )

        x = jnp.squeeze(x, axis=0)
        x = jnp.moveaxis(x, -1, 0)
        return x


class AvgPool1D(Pool):
    """One-dimensional downsample using an average over a sliding window."""

    def __init__(
        self,
        kernel_size,
        stride=None,
        padding=0,
        **kwargs,
    ):
        """**Arguments:**

        - `kernel_size`: The size of the convolutional kernel.
        - `stride`: The stride of the convolution.
        - `padding`: The amount of padding to apply before and after each
            spatial dimension.
        """

        super().__init__(
            init=0,
            operation=lax.add,
            num_spatial_dims=1,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            **kwargs,
        )

    def __call__(
        self, x: Array, *, key: Optional["jax.random.PRNGKey"] = None
    ) -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array of shape `(channels, dim)`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(channels, new_dim)`.
        """

        return super().__call__(x) / np.prod(self.kernel_size)


class MaxPool1D(Pool):
    """One-dimensional downsample using the maximum over a sliding window."""

    def __init__(
        self,
        kernel_size,
        stride=None,
        padding=0,
        **kwargs,
    ):
        """**Arguments:**

        - `kernel_size`: The size of the convolutional kernel.
        - `stride`: The stride of the convolution.
        - `padding`: The amount of padding to apply before and after each
            spatial dimension.
        """

        super().__init__(
            init=-jnp.inf,
            operation=lax.max,
            num_spatial_dims=1,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            **kwargs,
        )

    # Redefined to get them in the right order in docs
    def __call__(
        self, x: Array, *, key: Optional["jax.random.PRNGKey"] = None
    ) -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array of shape `(channels, dim)`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(channels, new_dim)`.
        """

        return super().__call__(x)


class AvgPool2D(Pool):
    """Two-dimensional downsample using an average over a sliding window."""

    def __init__(
        self,
        kernel_size,
        stride=None,
        padding=0,
        **kwargs,
    ):
        """**Arguments:**

        - `kernel_size`: The size of the convolutional kernel.
        - `stride`: The stride of the convolution.
        - `padding`: The amount of padding to apply before and after each
            spatial dimension.
        """

        super().__init__(
            init=0,
            operation=lax.add,
            num_spatial_dims=2,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            **kwargs,
        )

    def __call__(
        self, x: Array, *, key: Optional["jax.random.PRNGKey"] = None
    ) -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array of shape `(channels, dim_1, dim_2)`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(channels, new_dim_1, new_dim_2)`.
        """

        return super().__call__(x) / np.prod(self.kernel_size)


class MaxPool2D(Pool):
    """Two-dimensional downsample using the maximum over a sliding window."""

    def __init__(
        self,
        kernel_size,
        stride=None,
        padding=0,
        **kwargs,
    ):
        """**Arguments:**

        - `kernel_size`: The size of the convolutional kernel.
        - `stride`: The stride of the convolution.
        - `padding`: The amount of padding to apply before and after each
            spatial dimension.
        """

        super().__init__(
            init=-jnp.inf,
            operation=lax.max,
            num_spatial_dims=2,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            **kwargs,
        )

    # Redefined to get them in the right order in docs
    def __call__(
        self, x: Array, *, key: Optional["jax.random.PRNGKey"] = None
    ) -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array of shape `(channels, dim_1, dim_2)`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(channels, new_dim_1, new_dim_2)`.
        """

        return super().__call__(x)


class AvgPool3D(Pool):
    """Three-dimensional downsample using an average over a sliding window."""

    def __init__(
        self,
        kernel_size,
        stride=None,
        padding=0,
        **kwargs,
    ):
        """**Arguments:**

        - `kernel_size`: The size of the convolutional kernel.
        - `stride`: The stride of the convolution.
        - `padding`: The amount of padding to apply before and after each
            spatial dimension.
        """

        super().__init__(
            init=0,
            operation=lax.add,
            num_spatial_dims=3,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            **kwargs,
        )

    def __call__(
        self, x: Array, *, key: Optional["jax.random.PRNGKey"] = None
    ) -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array of shape
            `(channels, dim_1, dim_2, dim_3)`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(channels, new_dim_1, new_dim_2, new_dim_3)`.
        """

        return super().__call__(x) / np.prod(self.kernel_size)


class MaxPool3D(Pool):
    """Three-dimensional downsample using the maximum over a sliding window."""

    def __init__(
        self,
        kernel_size,
        stride=None,
        padding=0,
        **kwargs,
    ):
        """**Arguments:**

        - `kernel_size`: The size of the convolutional kernel.
        - `stride`: The stride of the convolution.
        - `padding`: The amount of padding to apply before and after each
            spatial dimension.
        """

        super().__init__(
            init=-jnp.inf,
            operation=lax.max,
            num_spatial_dims=3,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            **kwargs,
        )

    def __call__(
        self, x: Array, *, key: Optional["jax.random.PRNGKey"] = None
    ) -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array of shape
            `(channels, dim_1, dim_2, dim_3)`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(channels, new_dim_1, new_dim_2, new_dim_3)`.
        """

        return super().__call__(x)



from typing import Tuple

import jax
import jax.lax as lax
import jax.numpy as jnp
import jax.random as jr

from ..custom_types import Array
from ..module import Module, static_field
from .stateful import get_state, set_state, StateIndex


def _power_iteration(weight, u, v, eps):
    u = weight @ v
    u_norm = jnp.sqrt(jnp.sum(u**2))
    u = u / jnp.maximum(eps, u_norm)

    v = weight.T @ u
    v_norm = jnp.sqrt(jnp.sum(v**2))
    v = v / jnp.maximum(eps, v_norm)

    return u, v


class SpectralNorm(Module):
    """Applies spectral normalisation to a given parameter.

    Given a weight matrix $W$, and letting $σ(W)$ denote (an approximation to) its
    largest singular value, then this computes $W/σ(W)$.

    The approximation $σ(W)$ is computed using [power iterations](https://en.wikipedia.org/wiki/Power_iteration)
    which are updated (as a side-effect) every time $W/σ(W)$ is computed.

    Spectral normalisation is particularly commonly used when training generative
    adversarial networks; see [Spectral Normalization for Generative Adversarial Networks](https://arxiv.org/abs/1802.05957)
    for more details and motivation.

    [`equinox.experimental.SpectralNorm`][] should be used to replace an individual
    parameter. (Unlike some libraries, not the layer containing that parameter.)

    !!! example

        To add spectral normalisation during model creation:
        ```python
        import equinox as eqx
        import equinox.experimental as eqxe
        import jax.random as jr

        key = jr.PRNGKey(0)
        linear = eqx.nn.Linear(2, 2, key=key)
        sn_weight = eqxe.SpectralNorm(linear.weight, key=key)
        linear = eqx.tree_at(lambda l: l.weight, linear, sn_weight)
        ```

    !!! example

        Alternatively, iterate over the model to add spectral normalisation after model
        creation:
        ```python
        import equinox as eqx
        import equinox.experimental as eqxe
        import jax
        import jax.random as jr
        import functools as ft

        key = jr.PRNGKey(0)
        model_key, spectral_key = jr.split(key)
        SN = ft.partial(eqxe.SpectralNorm, key=spectral_key)

        def _is_linear(leaf):
            return isinstance(leaf, eqx.nn.Linear)

        def _apply_sn_to_linear(module):
            if _is_linear(module):
                module = eqx.tree_at(lambda m: m.weight, module, replace_fn=SN)
            return module

        def apply_sn(model):
            return jax.tree_map(_apply_sn_to_linear, model, is_leaf=_is_linear)

        model = eqx.nn.MLP(2, 2, 2, 2, key=model_key)
        model_with_sn = apply_sn(model)
        ```

    !!! example

        Switching the model to inference mode after training:
        ```python
        import equinox as eqx
        import equinox.experimental as eqxe
        import jax

        def _is_sn(leaf):
            return isinstance(leaf, eqxe.SpectralNorm)

        def _set_inference_on_sn(module):
            if _is_sn(module):
                module = eqx.tree_at(lambda m: m.inference, module, True)
            return module

        def set_inference(model):
            return jax.tree_map(_set_inference_on_sn, model, is_leaf=_is_sn)

        model = ...  # set up model, train it, etc.
        model = set_inference(model)
        ```

    !!! warning

        [`equinox.experimental.SpectralNorm`][] updates its running statistics as a side
        effect of its forward pass. Side effects are quite unusual in JAX; as such
        `SpectralNorm` is considered experimental. Let us know how you find it!
    """  # noqa: E501

    weight_shape: Tuple[int, ...] = static_field()
    weight: Array
    uv_index: StateIndex
    num_power_iterations: int = static_field()
    eps: float = static_field()
    inference: bool

    def __init__(
        self,
        weight: Array,
        num_power_iterations: int = 1,
        eps: float = 1e-12,
        inference: bool = False,
        *,
        key: "jax.random.PRNGKey",
        **kwargs
    ):
        """**Arguments:**

        - `weight`: The parameter (a JAX array) to apply spectral normalisation to.
        - `num_power_iterations`: The number of power iterations to apply every time the array is accessed.
        - `eps`: Epsilon for numerical stability when calculating norms.
        - `inference`: Whether this is in inference mode, at which time no power iterations are performed.
            This may be toggled with [`equinox.tree_inference`][].
        - `key`: A `jax.random.PRNGKey` used to provide randomness for initialisation. (Keyword only argument.)
        """
        super().__init__(**kwargs)
        if weight.ndim < 2:
            raise ValueError("`weight` must be at least two-dimensional")

        self.weight_shape = weight.shape
        weight = jnp.reshape(weight, (weight.shape[0], -1))
        self.weight = weight
        self.uv_index = StateIndex(inference=inference)
        self.num_power_iterations = num_power_iterations
        self.eps = eps
        self.inference = inference

        u_len, v_len = weight.shape
        ukey, vkey = jr.split(key)
        u0 = jr.normal(ukey, (u_len,))
        v0 = jr.normal(vkey, (v_len,))
        for _ in range(15):
            u0, v0 = _power_iteration(weight, u0, v0, eps)
        set_state(self.uv_index, (u0, v0))

    def __jax_array__(self):
        u_like = self.weight[:, -1]
        v_like = self.weight[-1]
        u, v = get_state(self.uv_index, (u_like, v_like))
        if not self.inference:
            for _ in range(self.num_power_iterations):
                u, v = _power_iteration(self.weight, u, v, self.eps)
            set_state(self.uv_index, (u, v))
        u = lax.stop_gradient(u)
        v = lax.stop_gradient(v)
        σ = jnp.einsum("i,ij,j->", u, self.weight, v)
        return jnp.reshape(self.weight / σ, self.weight_shape)

    @classmethod
    def withkey(cls, key: "jax.random.PRNGKey"):
        return lambda *args, **kwargs: cls(*args, key=key, **kwargs)



from .attention import MultiheadAttention
from .composed import MLP, Sequential
from .conv import (
    Conv,
    Conv1d,
    Conv2d,
    Conv3d,
    ConvTranspose,
    ConvTranspose1d,
    ConvTranspose2d,
    ConvTranspose3d,
)
from .dropout import Dropout
from .embedding import Embedding
from .linear import Identity, Linear
from .normalisation import GroupNorm, LayerNorm
from .pool import AvgPool1D, AvgPool2D, AvgPool3D, MaxPool1D, MaxPool2D, MaxPool3D, Pool
from .rnn import GRUCell, LSTMCell



import pathlib
from typing import Any, Callable, Union

import jax
import jax.numpy as jnp
import numpy as np

from . import experimental
from .custom_types import PyTree


def _default_serialise_filter_spec(f, x):
    if isinstance(x, jnp.ndarray):
        jnp.save(f, x)
    elif isinstance(x, np.ndarray):
        np.save(f, x)
    elif isinstance(x, (bool, float, complex, int)):
        np.save(f, x)
    elif isinstance(x, experimental.StateIndex):
        value, _, _ = x.unsafe_get()
        jnp.save(f, value)
    else:
        pass


def _default_deserialise_filter_spec(f, x):
    if isinstance(x, jnp.ndarray):
        return jnp.load(f)
    elif isinstance(x, np.ndarray):
        return np.load(f)
    elif isinstance(x, (bool, float, complex, int)):
        return np.load(f).item()
    elif isinstance(x, experimental.StateIndex):
        value = jnp.load(f)
        experimental.set_state(x, value)
        return x
    else:
        return x


def _with_suffix(path):
    path = pathlib.Path(path)
    if path.suffix == "":
        return path.with_suffix(".eqx")
    else:
        return path


def _assert_same(new, old):
    if type(new) is not type(old):
        raise RuntimeError(
            f"Deserialised leaf has changed type from {type(old)} in `like` to {type(new)} on disk."
        )
    if isinstance(new, (np.ndarray, jnp.ndarray)):
        if new.shape != old.shape:
            raise RuntimeError(
                f"Deserialised leaf has changed shape from {old.shape} in `like` to {new.shape} on disk."
            )
        if new.dtype != old.dtype:
            raise RuntimeError(
                f"Deserialised leaf has changed dtype from {old.dtype} in `like` to {new.dtype} on disk."
            )


def _is_index(x):
    return isinstance(x, experimental.StateIndex)


def tree_serialise_leaves(
    path: Union[str, pathlib.Path],
    pytree: PyTree,
    filter_spec=_default_serialise_filter_spec,
    is_leaf: Callable[[Any], bool] = _is_index,
) -> None:
    """Save the leaves of a PyTree to file.

    **Arguments:**

    - `path`: The file location to save values to.
    - `pytree`: The PyTree whose leaves will be saved.
    - `filter_spec`: Specifies how to save each kind of leaf. By default all JAX
        arrays, NumPy arrays, Python bool/int/float/complexes are saved,
        [`equinox.experimental.StateIndex`][] instances have their value looked up
        and saved, and all other leaf types are ignored.
    - `is_leaf`: Called on every node of `pytree`; if `True` then this node will be
        treated as a leaf.

    **Returns:**

    Nothing.

    !!! example

        This can be used to save a model to file.

        ```python
        import equinox as eqx
        import jax.random as jr

        model = eqx.nn.MLP(2, 2, 2, 2, key=jr.PRNGKey(0))
        eqx.tree_serialise_leaves("some_filename.eqx", model)
        ```

    !!! info

        `filter_spec` should typically be a function `(File, Any) -> None`, which takes
        a file handle and a leaf to save, and either saves the leaf to the file or
        does nothing.

        It can also be a PyTree of such functions, in which case the PyTree structure
        should be a prefix of `pytree`, and each function will be mapped over the
        corresponding sub-PyTree of `pytree`.
    """

    with open(_with_suffix(path), "wb") as f:

        def _serialise(spec, x):
            def __serialise(y):
                spec(f, y)

            jax.tree_map(__serialise, x, is_leaf=is_leaf)

        jax.tree_map(_serialise, filter_spec, pytree)


def tree_deserialise_leaves(
    path: Union[str, pathlib.Path],
    like: PyTree,
    filter_spec=_default_deserialise_filter_spec,
    is_leaf: Callable[[Any], bool] = _is_index,
) -> PyTree:
    """Load the leaves of a PyTree from a file.

    **Arguments:**

    - `path`: The file location to load values from.
    - `like`: A PyTree of the same structure, and with leaves of the same type, as the
        PyTree being loaded. Those leaves which are loaded will replace the
        corresponding leaves of `like`.
    - `filter_spec`: Specifies how to load each kind of leaf. By default all JAX
        arrays, NumPy arrays, Python bool/int/float/complexes are loaded, and
        [`equinox.experimental.StateIndex`][] instances have their value looked up
        and stored, and all other leaf types are not loaded (and will retain their
        value from `like`).
    - `is_leaf`: Called on every node of `like`; if `True` then this node will be
        treated as a leaf.

    **Returns:**

    The loaded PyTree, formed by iterating over `like` and replacing some of its leaves
    with the leaves saved in `path`.

    !!! example

        This can be used to load a model from file.

        ```python
        import equinox as eqx
        import jax.random as jr

        model = eqx.nn.MLP(2, 2, 2, 2, key=jr.PRNGKey(0))
        eqx.tree_serialise_leaves("some_filename.eqx", model)
        model2 = eqx.tree_deserialise_leaves("some_filename.eqx", model)
        ```

    !!! info

        `filter_spec` should typically be a function `(File, Any) -> Any`, which takes
        a file handle and a leaf from `like`, and either returns the corresponding
        loaded leaf, or retuns the leaf from `like` unchanged.

        It can also be a PyTree of such functions, in which case the PyTree structure
        should be a prefix of `pytree`, and each function will be mapped over the
        corresponding sub-PyTree of `pytree`.
    """

    with open(_with_suffix(path), "rb") as f:

        def _deserialise(spec, x):
            def __deserialise(y):
                return spec(f, y)

            return jax.tree_map(__deserialise, x, is_leaf=is_leaf)

        out = jax.tree_map(_deserialise, filter_spec, like)
    jax.tree_map(_assert_same, out, like, is_leaf=is_leaf)
    return out



from typing import Any, Callable, Optional, Sequence, Union

import jax
import jax.numpy as jnp
import numpy as np

from .custom_types import PyTree, sentinel
from .doc_utils import doc_repr


_Node = doc_repr(Any, "Node")


class _LeafWrapper:
    def __init__(self, value: Any):
        self.value = value


def _remove_leaf_wrapper(x: _LeafWrapper) -> Any:
    assert type(x) is _LeafWrapper
    return x.value


class _CountedIdDict:
    def __init__(self, keys, values):
        assert len(keys) == len(values)
        self._dict = {id(k): v for k, v in zip(keys, values)}
        self._count = {id(k): 0 for k in keys}

    def __contains__(self, item):
        return id(item) in self._dict

    def __getitem__(self, item):
        self._count[id(item)] += 1
        return self._dict[id(item)]

    def get(self, item, default):
        try:
            return self[item]
        except KeyError:
            return default

    def count(self, item):
        return self._count[id(item)]


def tree_at(
    where: Callable[[PyTree], Union[_Node, Sequence[_Node]]],
    pytree: PyTree,
    replace: Union[Any, Sequence[Any]] = sentinel,
    replace_fn: Callable[[_Node], Any] = sentinel,
    is_leaf: Optional[Callable[[Any], bool]] = None,
):
    """Updates a PyTree out-of-place; a bit like using `.at[].set()` on a JAX array.

    **Arguments:**

    - `where`: A callable `PyTree -> Node` or `PyTree -> Sequence[Node]`. It should
        consume a PyTree with the same structure as `pytree`, and return the node or
        nodes that should be replaced. For example
        `where = lambda mlp: mlp.layers[-1].linear.weight`.
    - `pytree`: The PyTree to modify.
    - `replace`: Either a single element, or a sequence of the same length as returned
        by `where`. This specifies the replacements to make at the locations specified
        by `where`. Mutually exclusive with `replace_fn`.
    - `replace_fn`: A function `Node -> Any`. It will be called on every node specified
        by `where`. The return value from `replace_fn` will be used in its place.
        Mutually exclusive with `replace`.
    - `is_leaf`: As `jax.tree_flatten`. For example pass `is_leaf=lambda x: x is None`
        to be able to replace `None` values using `tree_at`.

    **Returns:**

    A copy of the input PyTree, with the appropriate modifications.

    !!! example

        This can be used to help specify the weights of a model to train or not to
        train. For example the following will train only the weight of the final linear
        layer of an MLP:

        ```python
        def loss(model, ...):
            ...

        model = eqx.nn.MLP(...)
        trainable = jax.tree_map(lambda _: False, model)
        trainable = equinox.tree_at(lambda mlp: mlp.layers[-1].linear.weight, model, replace=True)
        grad_loss = equinox.filter_grad(loss, arg=trainable)
        grads = grad_loss(model)
        ```
    """

    # We need to specify a particular node in a PyTree.
    # This is surprisingly difficult to do! As far as I can see, pretty much the only
    # way of doing this is to specify e.g. `x.foo[0].bar` via `is`, and then pulling
    # a few tricks to try and ensure that the same object doesn't appear multiple
    # times in the same PyTree.
    #
    # So this first `tree_map` serves a dual purpose.
    # 1) Makes a copy of the composite nodes in the PyTree, to avoid aliasing via
    #    e.g. `pytree=[(1,)] * 5`. This has the tuple `(1,)` appear multiple times.
    # 2) It makes each leaf be a unique Python object, as it's wrapped in
    #    `_LeafWrapper`. This is needed because Python caches a few builtin objects:
    #    `assert 0 + 1 is 1`. I think only a few leaf types are subject to this.
    # So point 1) should ensure that all composite nodes are unique Python objects,
    # and point 2) should ensure that all leaves are unique Python objects.
    # Between them, all nodes of `pytree` are handled.
    #
    # I think pretty much the only way this can fail is when using a custom node with
    # singleton-like flatten+unflatten behaviour, which is pretty edge case. And we've
    # added a check for it at the bottom of this function, just to be sure.
    #
    # Whilst we're here: we also double-check that `where` is well-formed and doesn't
    # use leaf information. (As else `node_or_nodes` will be wrong.)
    node_or_nodes_nowrapper = where(pytree)
    pytree = jax.tree_map(_LeafWrapper, pytree, is_leaf=is_leaf)
    node_or_nodes = where(pytree)
    leaves1, structure1 = jax.tree_flatten(node_or_nodes_nowrapper, is_leaf=is_leaf)
    leaves2, structure2 = jax.tree_flatten(node_or_nodes)
    leaves2 = [_remove_leaf_wrapper(x) for x in leaves2]
    if (
        structure1 != structure2
        or len(leaves1) != len(leaves2)
        or any(l1 is not l2 for l1, l2 in zip(leaves1, leaves2))
    ):
        raise ValueError(
            "`where` must use just the PyTree structure of `pytree`. `where` must not "
            "depend on the leaves in `pytree`."
        )
    del node_or_nodes_nowrapper, leaves1, structure1, leaves2, structure2

    # Normalise whether we were passed a single node or a sequence of nodes.
    in_pytree = False

    def _in_pytree(x):
        nonlocal in_pytree
        if x is node_or_nodes:  # noqa: F821
            in_pytree = True

    jax.tree_map(_in_pytree, pytree, is_leaf=lambda x: x is node_or_nodes)  # noqa: F821
    if in_pytree:
        nodes = (node_or_nodes,)
        if replace is not sentinel:
            replace = (replace,)
    else:
        nodes = node_or_nodes
    del in_pytree, node_or_nodes

    # Normalise replace vs replace_fn
    if replace is sentinel:
        if replace_fn is sentinel:
            raise ValueError(
                "Precisely one of `replace` and `replace_fn` must be specified."
            )
        else:

            def _replace_fn(x):
                x = jax.tree_map(_remove_leaf_wrapper, x)
                return replace_fn(x)

            replace_fns = [_replace_fn] * len(nodes)
    else:
        if replace_fn is sentinel:
            if len(nodes) != len(replace):
                raise ValueError(
                    "`where` must return a sequence of leaves of the same length as "
                    "`replace`."
                )
            replace_fns = [lambda _, r=r: r for r in replace]
        else:
            raise ValueError(
                "Precisely one of `replace` and `replace_fn` must be specified."
            )
    node_replace_fns = _CountedIdDict(nodes, replace_fns)

    # Actually do the replacement
    def _make_replacement(x: _Node) -> Any:
        return node_replace_fns.get(x, _remove_leaf_wrapper)(x)

    out = jax.tree_map(
        _make_replacement, pytree, is_leaf=lambda x: x in node_replace_fns
    )

    # Check that `where` is well-formed.
    for node in nodes:
        count = node_replace_fns.count(node)
        if count == 0:
            raise ValueError(
                "`where` does not specify an element or elements of `pytree`."
            )
        elif count == 1:
            pass
        else:
            raise ValueError(
                "`where` does not uniquely identify a single element of `pytree`. This "
                "usually occurs when trying to replace a `None` value:\n"
                "\n"
                "  >>> eqx.tree_at(lambda t: t[0], (None, None, 1), True)\n"
                "\n"
                "\n"
                "for which the fix is to specify that `None`s should be treated as "
                "leaves:\n"
                "\n"
                "  >>> eqx.tree_at(lambda t: t[0], (None, None, 1), True,\n"
                "  ...             is_leaf=lambda x: x is None)"
            )

    return out


def tree_equal(*pytrees: PyTree) -> bool:
    """Returns `True` if all input PyTrees are equal. Every PyTree must have the same
    structure. Any JAX or NumPy arrays (as leaves) must have the same shape, dtype, and
    values to be considered equal. JAX arrays and NumPy arrays are not considered equal
    to each other.

    **Arguments:**

    - `*pytrees`: Any number of PyTrees each with any structure.

    **Returns:**

    A boolean.
    """
    flat, treedef = jax.tree_flatten(pytrees[0])
    array_types = (jnp.ndarray, np.ndarray)
    out = True
    for pytree in pytrees[1:]:
        flat_, treedef_ = jax.tree_flatten(pytree)
        if treedef_ != treedef:
            return False
        for elem, elem_ in zip(flat, flat_):
            if isinstance(elem, array_types):
                if isinstance(elem_, array_types):
                    if (
                        (type(elem) != type(elem_))
                        or (elem.shape != elem_.shape)
                        or (elem.dtype != elem_.dtype)
                    ):
                        return False
                    allsame = (elem == elem_).all()
                    if allsame is False:
                        return False
                    out = out & allsame
                else:
                    return False
            else:
                if isinstance(elem_, array_types):
                    return False
                else:
                    if elem != elem_:
                        return False
    return out


def _has_inference(leaf):
    return hasattr(leaf, "inference")


def _inferences(pytree):
    return tuple(
        x.inference
        for x in jax.tree_leaves(pytree, is_leaf=_has_inference)
        if _has_inference(x)
    )


def tree_inference(pytree: PyTree, value: bool) -> PyTree:
    """Convenience function for setting all `inference` attributes on a PyTree.

    Equivalent to:
    ```python
    has_inference = lambda leaf: hasattr(leaf, "inference")

    def where(pytree):
        return tuple(x.inference
                     for x in jax.tree_leaves(pytree, is_leaf=has_inference)
                     if has_inference(x))

    equinox.tree_at(where, pytree, replace_fn=lambda _: value)
    ```

    `inference` flags are used to toggle the behaviour of a number of the pre-built
    neural network layers, such as [`equinox.nn.Dropout`][] or
    [`equinox.experimental.BatchNorm`][].

    **Arguments:**

    - `pytree`: the PyTree to modify.
    - `value`: the value to set all `inference` attributes to.

    **Returns:**

    A copy of `pytree` with all `inference` flags set to `value`.
    """

    # For the sake of equinox.experimental.StateIndex. This won't defend against anyone
    # setting inference flags manually using tree_at etc., but it should help overall.
    if isinstance(jnp.array(0) + 1, jax.core.Tracer):
        raise RuntimeError(
            "inference flags should not be set whilst jit'ing, vmap'ing etc."
        )

    return tree_at(_inferences, pytree, replace_fn=lambda _: value)



from typing import Tuple

import jax.numpy as jnp

from ..custom_types import Array


def left_broadcast_to(arr: Array, shape: Tuple[int]):
    arr = arr.reshape(arr.shape + (1,) * (len(shape) - arr.ndim))
    return jnp.broadcast_to(arr, shape)
